@@ -1,13 +1,27 @@
 """Sqlutilpy module to access SQL databases
 """
 from __future__ import print_function
-import types
 import numpy
-import sys
 import numpy as np
-import time
 import psycopg2
 import threading
+import collections
+import warnings
+from numpy.core import numeric as sb
+from select import select
+from psycopg2.extensions import POLL_OK, POLL_READ, POLL_WRITE
+
+try:
+    import astropy.table as atpy
+except ImportError:
+    # astropy is not installed
+    atpy = None
+try:
+    import pandas
+except ImportError:
+    # pandas is not installed
+    pandas = None
+
 try:
     from cStringIO import StringIO
 except ImportError:
@@ -15,41 +29,36 @@ except ImportError:
     from io import BytesIO as StringIO
 try:
     import queue
-except:
+except ImportError:
     import Queue as queue
-import collections
-import warnings
-try:
-    dictclass = collections.OrderedDict
-except AttributeError:
-    try:
-        import ordereddict
-        dictclass = ordereddict.OrderedDict
-    except ImportError:
-        warnings.warn('No ordered dict library found')
-        dictclass = dict
-from numpy.core import numeric as sb
-from numpy.core import numerictypes as nt
-from select import select
-from psycopg2.extensions import POLL_OK, POLL_READ, POLL_WRITE
+
+_WAIT_SELECT_TIMEOUT = 10
+
+
+class config:
+    arraysize = 100000
+
+
+class SqlUtilException(Exception):
+    pass
 
 
 def __wait_select_inter(conn):
     """ Make the queries interruptable by Ctrl-C
 
-    Taken from http://initd.org/psycopg/articles/2014/07/20/cancelling-postgresql-statements-python/"""
+    Taken from http://initd.org/psycopg/articles/2014/07/20/cancelling-postgresql-statements-python/
+    """
     while True:
         try:
             state = conn.poll()
             if state == POLL_OK:
                 break
             elif state == POLL_READ:
-                select([conn.fileno()], [], [])
+                select([conn.fileno()], [], [], _WAIT_SELECT_TIMEOUT)
             elif state == POLL_WRITE:
-                select([], [conn.fileno()], [])
+                select([], [conn.fileno()], [], _WAIT_SELECT_TIMEOUT)
             else:
-                raise conn.OperationalError(
-                    "bad state from poll: %s" % state)
+                raise conn.OperationalError("bad state from poll: %s" % state)
         except KeyboardInterrupt:
             conn.cancel()
             # the loop will be broken by a server error
@@ -59,8 +68,13 @@ def __wait_select_inter(conn):
 psycopg2.extensions.set_wait_callback(__wait_select_inter)
 
 
-def getConnection(db=None, driver=None, user=None,
-                  password=None, host=None, port=5432, timeout=None):
+def getConnection(db=None,
+                  driver=None,
+                  user=None,
+                  password=None,
+                  host=None,
+                  port=5432,
+                  timeout=None):
     """ Retrieve the connection to the DB object
 
     Parameters
@@ -89,7 +103,7 @@ def getConnection(db=None, driver=None, user=None,
 
     """
     if driver == 'psycopg2':
-        conn_str = "dbname=%s host=%s port=%d" % (db, host, port)
+        conn_str = "dbname=%s host=%s port=%d" % (db, host, (port or 5432))
         if user is not None:
             conn_str = conn_str + ' user=%s' % user
         if password is not None:
@@ -118,9 +132,11 @@ def getCursor(conn, driver=None, preamb=None, notNamed=False):
         if notNamed:
             return cur
         cur = conn.cursor(name='sqlutilcursor')
-        cur.arraysize = 100000
+        cur.arraysize = config.arraysize
     elif driver == 'sqlite3':
         cur = conn.cursor()
+        if preamb is not None:
+            cur.execute(preamb)
     return cur
 
 
@@ -129,13 +145,12 @@ def __fromrecords(recList, dtype=None, intNullVal=None):
                     support conversion null integers to intNullVal
     """
 
-    nfields = len(recList[0])
     shape = None
     descr = sb.dtype((np.core.records.record, dtype))
     try:
         retval = sb.array(recList, dtype=descr)
     except TypeError:  # list of lists instead of list of tuples
-        shape = (len(recList),)
+        shape = (len(recList), )
         _array = np.core.records.recarray(shape, descr)
         try:
             for k in range(_array.size):
@@ -151,7 +166,8 @@ def __fromrecords(recList, dtype=None, intNullVal=None):
                     convs.append(lambda x: x)
             convs = tuple(convs)
 
-            def convF(x): return [convs[_](x[_]) for _ in range(ncols)]
+            def convF(x):
+                return [convs[_](x[_]) for _ in range(ncols)]
 
             for k in range(k, _array.size):
                 try:
@@ -170,7 +186,7 @@ def __fromrecords(recList, dtype=None, intNullVal=None):
 
 def __converter(qIn, qOut, endEvent, dtype, intNullVal):
     """ Convert the input stream of tuples into numpy arrays """
-    while(not endEvent.is_set()):
+    while (not endEvent.is_set()):
         try:
             tups = qIn.get(True, 0.1)
         except queue.Empty:
@@ -184,8 +200,8 @@ def __converter(qIn, qOut, endEvent, dtype, intNullVal):
         qOut.put(res)
 
 
-def getDType(row, typeCodes, strLength):
-    __pgTypeHash = {
+def __getDType(row, typeCodes, strLength):
+    pgTypeHash = {
         16: bool,
         18: str,
         20: 'i8',
@@ -195,42 +211,64 @@ def getDType(row, typeCodes, strLength):
         25: '|U%d',
         700: 'f4',
         701: 'f8',
+        1000: bool,
+        1005: 'i2',
+        1007: 'i4',
+        1016: 'i8',
+        1021: 'f4',
+        1022: 'f8',
         1042: '|U%d',  # character()
         1043: '|U%d',  # varchar
-        1700: 'f8',	 # numeric
+        1700: 'f8',  # numeric
         1114: '<M8[us]',  # timestamp
-        1082: '<M8[us]'	 # date
+        1082: '<M8[us]'  # date
     }
-    strTypes = [25,1042,1043]
-    
-    pgTypes=[]
-    
-    for i,(curv, curt) in enumerate(zip(row, typeCodes)):
-        if curt not in __pgTypeHash:
-            raise Exception('Unknown PG type  %d'%curt)
-        pgType =__pgTypeHash[curt]
+    strTypes = [25, 1042, 1043]
+
+    pgTypes = []
+
+    for i, (curv, curt) in enumerate(zip(row, typeCodes)):
+        if curt not in pgTypeHash:
+            raise SqlUtilException('Unknown PG type  %d' % curt)
+        pgType = pgTypeHash[curt]
         if curt in strTypes:
             if curv is not None:
+                # if the first string value is longer than
+                # strLength use that as a maximum
                 curmax = max(strLength, len(curv))
             else:
+                # if the first value is null
+                # just use strLength
                 curmax = strLength
-            pgType = pgType %(curmax,)
+            pgType = pgType % (curmax, )
         if curt not in strTypes:
             try:
                 len(curv)
                 pgType = 'O'
             except TypeError:
                 pass
-        pgTypes.append(('a%d'%i, pgType))
+        pgTypes.append(('a%d' % i, pgType))
     dtype = numpy.dtype(pgTypes)
     return dtype
-        
-def get(query, params=None, db="wsdb", driver="psycopg2", user=None,
-        password=None, host='localhost', preamb=None,
-        conn=None, port=5432,
-        strLength=10, timeout=None, notNamed=False,
-        asDict=False, intNullVal=-9999):
-    '''Executes the sql query and returns the tuple or dictionary with the numpy arrays.
+
+
+def get(query,
+        params=None,
+        db="wsdb",
+        driver="psycopg2",
+        user=None,
+        password=None,
+        host='localhost',
+        preamb=None,
+        conn=None,
+        port=5432,
+        strLength=10,
+        timeout=None,
+        notNamed=False,
+        asDict=False,
+        intNullVal=-9999):
+    '''Executes the sql query and returns the tuple or dictionary
+    with the numpy arrays.
 
     Parameters
     ----------
@@ -242,7 +280,8 @@ def get(query, params=None, db="wsdb", driver="psycopg2", user=None,
     conn : object
           The connection object to the DB (optional) to avoid reconnecting
     asDict : boolean
-          Flag whether to retrieve the results as a dictionary with column names as keys
+        Flag whether to retrieve the results as a dictionary with column
+        names as keys
     strLength : integer
          The maximum length of the string.
         Strings will be truncated to this length
@@ -267,7 +306,8 @@ def get(query, params=None, db="wsdb", driver="psycopg2", user=None,
     Returns
     -------
     ret : Tuple or dictionary
-        By default you get a tuple with numpy arrays for each column in your query.
+        By default you get a tuple with numpy arrays for each column
+        in your query.
         If you specified asDict keyword then you get an ordered dictionary with
         your columns.
 
@@ -281,8 +321,13 @@ def get(query, params=None, db="wsdb", driver="psycopg2", user=None,
     '''
     connSupplied = (conn is not None)
     if not connSupplied:
-        conn = getConnection(db=db, driver=driver, user=user, password=password,
-                             host=host, port=port, timeout=timeout)
+        conn = getConnection(db=db,
+                             driver=driver,
+                             user=user,
+                             password=password,
+                             host=host,
+                             port=port,
+                             timeout=timeout)
     try:
         cur = getCursor(conn, driver=driver, preamb=preamb, notNamed=notNamed)
 
@@ -301,38 +346,53 @@ def get(query, params=None, db="wsdb", driver="psycopg2", user=None,
         colNames = []
         if driver == 'psycopg2':
             try:
-                while(True):
+                while (True):
+                    # Iterating over the cursor, retrieving batches of results
+                    # and then sending them for conversion
                     tups = cur.fetchmany()
                     desc = cur.description
 
+                    # No more data
                     if tups == []:
                         break
-                    
+
+                    # Send the new batch for conversion
                     qIn.put(tups)
+
+                    # If the is just the start we need to launch the
+                    # thread doing the conversion
                     if nrec == 0:
                         typeCodes = [_tmp.type_code for _tmp in desc]
                         colNames = [_tmp.name for _tmp in cur.description]
-                        dtype = getDType(tups[0], typeCodes, strLength)
-                        proc = threading.Thread(target=__converter, args=(
-                            qIn, qOut, endEvent, dtype, intNullVal))
+                        dtype = __getDType(tups[0], typeCodes, strLength)
+                        proc = threading.Thread(target=__converter,
+                                                args=(qIn, qOut, endEvent,
+                                                      dtype, intNullVal))
                         proc.start()
 
+                    # nrec is the number of batches in conversion currently
                     nrec += 1
+
+                    # Try to retrieve one processed batch without waiting
+                    # on it
                     try:
                         reslist.append(qOut.get(False))
                         nrec -= 1
                     except queue.Empty:
                         pass
-                try:
-                    while(nrec != 0):
-                        try:
-                            reslist.append(qOut.get(True, 0.1))
-                            nrec -= 1
-                        except:
-                            if endEvent.is_set():
-                                raise Exception('Child thread failed')
-                except queue.Empty:
-                    pass
+
+                # Now we are done fetching the data from the DB, we
+                # just need to assemble the converted results
+                while (nrec != 0):
+                    try:
+                        reslist.append(qOut.get(True, 0.1))
+                        nrec -= 1
+                    except queue.Empty:
+                        # continue looping unless the endEvent was set
+                        # which should happen in the case of the crash
+                        # of the converter thread
+                        if endEvent.is_set():
+                            raise Exception('Child thread failed')
                 endEvent.set()
             except BaseException:
                 endEvent.set()
@@ -348,47 +408,29 @@ def get(query, params=None, db="wsdb", driver="psycopg2", user=None,
             if reslist == []:
                 nCols = len(desc)
                 res = numpy.array([],
-                                  dtype=numpy.dtype(
-                    [('a%d' % i, 'f') for i in range(nCols)])
-                )
+                                  dtype=numpy.dtype([('a%d' % i, 'f')
+                                                     for i in range(nCols)]))
             else:
                 res = numpy.concatenate(reslist)
 
         elif driver == 'sqlite3':
             tups = cur.fetchall()
+            colNames = [_tmp[0] for _tmp in cur.description]
             if len(tups) > 0:
-                _cast = {types.BooleanType: numpy.bool,
-                         types.IntType: numpy.int32,
-                         types.LongType: numpy.int64,
-                         types.FloatType: numpy.float64,
-                         types.StringType: numpy.str,
-                         types.UnicodeType: numpy.str}
-                try:
-                    typelist = [_cast[type(tmp)] for tmp in tups[0]]
-                except KeyError:
-                    raise Exception("Unknown datatype")
                 res = numpy.core.records.array(tups)
             else:
-                return None
+                return [[]] * len(cur.description)
 
         res = [res[tmp] for tmp in res.dtype.names]
 
     except BaseException:
-        try:
-            conn.rollback()
-        except:
-            pass
-        if not connSupplied:
-            try:
-                conn.close()  # do not close if we were given the connection
-            except:
-                pass
+        failure_cleanup(conn, connSupplied)
         raise
 
     cur.close()
 
     if asDict:
-        resDict = dictclass()
+        resDict = collections.OrderedDict()
         repeats = {}
         for _n, _v in zip(colNames, res):
             if _n in resDict:
@@ -404,9 +446,16 @@ def get(query, params=None, db="wsdb", driver="psycopg2", user=None,
     return res
 
 
-def execute(query, params=None, db='wsdb', driver="psycopg2", user=None,
-            password=None, host='localhost',
-            conn=None, preamb=None, timeout=None,
+def execute(query,
+            params=None,
+            db='wsdb',
+            driver="psycopg2",
+            user=None,
+            password=None,
+            host='localhost',
+            conn=None,
+            preamb=None,
+            timeout=None,
             noCommit=False):
     """Execute a given SQL command without returning the results
 
@@ -422,21 +471,25 @@ def execute(query, params=None, db='wsdb', driver="psycopg2", user=None,
     driver : string
         Driver for the DB connection ('psucopg2' or 'sqlite3')
     user : string, optional
-          user name for the DB connection
+        user name for the DB connection
     password : string, optional
-         DB connection password
+        DB connection password
     host : string, optional
-         Hostname of the database
+        Hostname of the database
     port : integer, optional
-         Port of the database
+        Port of the database
     noCommit: bool
-         By default execute() will commit your command. If you say noCommit, the
-         commit won't be issued.
+        By default execute() will commit your command.
+        If you say noCommit, the commit won't be issued.
     """
     connSupplied = (conn is not None)
     if not connSupplied:
-        conn = getConnection(db=db, driver=driver, user=user, password=password,
-                             host=host, timeout=timeout)
+        conn = getConnection(db=db,
+                             driver=driver,
+                             user=user,
+                             password=password,
+                             host=host,
+                             timeout=timeout)
     try:
         cur = getCursor(conn, driver=driver, preamb=preamb, notNamed=True)
         if params is not None:
@@ -445,15 +498,7 @@ def execute(query, params=None, db='wsdb', driver="psycopg2", user=None,
             # sqlite3 doesn't like params here...
             cur.execute(query)
     except BaseException:
-        try:
-            conn.rollback()
-        except Exception:
-            pass
-        if not connSupplied:
-            try:
-                conn.close()  # do not close if we were given the connection
-            except:
-                pass
+        failure_cleanup(conn, connSupplied)
         raise
     cur.close()
     if not noCommit:
@@ -463,13 +508,12 @@ def execute(query, params=None, db='wsdb', driver="psycopg2", user=None,
 
 
 def __create_schema(tableName, arrays, names, temp=False):
-    hash = dict([
-        (np.int32, 'integer'),
-        (np.int64, 'bigint'),
-        (np.float32, 'real'),
-        (np.float64, 'double precision'),
-        (np.string_, 'varchar'),
-    ])
+    hash = dict([(np.int32, 'integer'), (np.int64, 'bigint'),
+                 (np.uint64, 'bigint'), (np.int16, 'smallint'),
+                 (np.uint8, 'bigint'), (np.float32, 'real'),
+                 (np.float64, 'double precision'), (np.string_, 'varchar'),
+                 (np.str_, 'varchar'), (np.bool_, 'boolean'),
+                 (np.datetime64, 'timestamp')])
     if temp:
         temp = 'temporary'
     else:
@@ -481,32 +525,53 @@ def __create_schema(tableName, arrays, names, temp=False):
     return outp + '(' + ','.join(outp1) + ')'
 
 
-def __print_arrays(arrays, f):
-    hash = dict([
-        (np.int32, '%d'),
-        (np.int64, '%d'),
-        (np.float32, '%.18e'),
-        (np.float64, '%.18e'),
-        (np.string_, '%s')
-    ])
+def __print_arrays(arrays, f, sep=' '):
+    hash = dict([(np.int32, '%d'), (np.int64, '%d'), (np.int16, '%d'),
+                 (np.uint8, '%d'), (np.float32, '%.18e'),
+                 (np.float64, '%.18e'), (np.string_, '%s'), (np.str_, '%s'),
+                 (np.datetime64, '%s'), (np.bool_, '%d')])
     fmt = [hash[x.dtype.type] for x in arrays]
     recarr = np.rec.fromarrays(arrays)
-    np.savetxt(f, recarr, fmt=fmt)
+    np.savetxt(f, recarr, fmt=fmt, delimiter=sep)
 
 
-def upload(tableName, arrays, names, db="wsdb", driver="psycopg2", user=None,
-           password=None, host='locahost',
-           conn=None, preamb=None, timeout=None,
-           noCommit=False, temp=False,
-           analyze=False, createTable=True):
+def failure_cleanup(conn, connSupplied):
+    try:
+        conn.rollback()
+    except Exception:
+        pass
+    if not connSupplied:
+        try:
+            conn.close()  # do not close if we were given the connection
+        except:
+            pass
+
+
+def upload(tableName,
+           arrays,
+           names=None,
+           db="wsdb",
+           driver="psycopg2",
+           user=None,
+           password=None,
+           host='locahost',
+           conn=None,
+           preamb=None,
+           timeout=None,
+           noCommit=False,
+           temp=False,
+           analyze=False,
+           createTable=True):
     """ Upload the data stored in the tuple of arrays in the DB
 
     Parameters
     ----------
     tableName : string
         The name of the table where the data will be uploaded
-    arrays : tuple
+    arrays_or_table : tuple
         Tuple of arrays thar will be columns of the new table
+        If names are not specified, I this parameter can be pandas or
+        astropy table
     names : tuple
         Tuple of strings with column names
 
@@ -516,35 +581,67 @@ def upload(tableName, arrays, names, db="wsdb", driver="psycopg2", user=None,
     >>> y = x**.5
     >>> sqlutil.upload('mytable',(x,y),('xcol','ycol'))
     """
-    arrays = [np.asarray(_) for _ in arrays]
     connSupplied = (conn is not None)
+    sep = '|'
     if not connSupplied:
-        conn = getConnection(db=db, driver=driver, user=user, password=password,
-                             host=host, timeout=timeout)
+        conn = getConnection(db=db,
+                             driver=driver,
+                             user=user,
+                             password=password,
+                             host=host,
+                             timeout=timeout)
+    if names is None:
+        for i in range(1):
+            # we assume that we were given a table
+            if atpy is not None:
+                if isinstance(arrays, atpy.Table):
+                    names = arrays.columns
+                    arrays = [arrays[_] for _ in names]
+                    break
+            if pandas is not None:
+                if isinstance(arrays, pandas.DataFrame):
+                    names = arrays.columns
+                    arrays = [arrays[_] for _ in names]
+                    break
+            if isinstance(arrays, dict):
+                names = arrays.keys()
+                arrays = [arrays[_] for _ in names]
+                break
+            if names is None:
+                raise Exception('you either need to give astropy \
+table/pandas/dictionary or provide a separate list of arrays and their names')
+
+    arrays = [np.asarray(_) for _ in arrays]
+    repl_char = {' ': '_', '-': '_', '(': '_', ')': '_', '[': '_', ']': '_'}
+    fixed_names = []
+    for name in names:
+        fixed_name = name + ''
+        for k in repl_char.keys():
+            fixed_name = fixed_name.replace(k, repl_char[k])
+        if fixed_name != name:
+            warnings.warn('''Renamed column '%s' to '%s' ''' %
+                          (name, fixed_name))
+        fixed_names.append(fixed_name)
+    names = fixed_names
     try:
         cur = getCursor(conn, driver=driver, preamb=preamb, notNamed=True)
         if createTable:
             query1 = __create_schema(tableName, arrays, names, temp=temp)
             cur.execute(query1)
-        f = StringIO()
-        __print_arrays(arrays, f)
-        f.seek(0)
-        try:
-            thread = psycopg2.extensions.get_wait_callback()
-            psycopg2.extensions.set_wait_callback(None)
-            cur.copy_from(f, tableName, sep=' ', columns=names)
-        finally:
-            psycopg2.extensions.set_wait_callback(thread)
-    except BaseException:
-        try:
-            conn.rollback()
-        except Exception:
-            pass
-        if not connSupplied:
+        nsplit = 100000
+        N = len(arrays[0])
+        for i in range(0, N, nsplit):
+            f = StringIO()
+            __print_arrays([_[i:i + nsplit] for _ in arrays], f, sep=sep)
+            f.seek(0)
             try:
-                conn.close()  # do not close if we were given the connection
-            except:
-                pass
+                thread = psycopg2.extensions.get_wait_callback()
+                psycopg2.extensions.set_wait_callback(None)
+                cur.copy_from(f, tableName, sep=sep, columns=names)
+            finally:
+                psycopg2.extensions.set_wait_callback(thread)
+    except BaseException:
+        failure_cleanup(conn, connSupplied)
         raise
     if analyze:
         cur.execute('analyze %s' % tableName)
@@ -555,11 +652,21 @@ def upload(tableName, arrays, names, db="wsdb", driver="psycopg2", user=None,
         conn.close()  # do not close if we were given the connection
 
 
-def local_join(query, tableName, arrays, names, db="wsdb", driver="psycopg2", user=None,
-               password=None, host='locahost',
+def local_join(query,
+               tableName,
+               arrays,
+               names,
+               db=None,
+               driver="psycopg2",
+               user=None,
+               password=None,
+               host='locahost',
                port=5432,
-               conn=None, preamb=None, timeout=None,
-               strLength=20, asDict=False):
+               conn=None,
+               preamb=None,
+               timeout=None,
+               strLength=20,
+               asDict=False):
     """ Join the data from python with the data in the database
     This command first uploads the data in the DB and then runs a
     user specified query.
@@ -575,18 +682,32 @@ def local_join(query, tableName, arrays, names, db="wsdb", driver="psycopg2", us
     --------
     >>> x = np.arange(10)
     >>> y = x**.5
-    >>> sqlutil.local_join('select * from mytable as m, sometable as s where s.id=m.xcol',
-                                                                                                    'mytable',(x,y),('xcol','ycol'))
+    >>> sqlutil.local_join('select * from mytable as m, sometable as s
+        where s.id=m.xcol',
+        'mytable',(x,y),('xcol','ycol'))
     """
 
     connSupplied = (conn is not None)
     if not connSupplied:
-        conn = getConnection(db=db, driver=driver, user=user, password=password,
-                             host=host, timeout=timeout, port=port)
+        conn = getConnection(db=db,
+                             driver=driver,
+                             user=user,
+                             password=password,
+                             host=host,
+                             timeout=timeout,
+                             port=port)
 
-    upload(tableName, arrays, names, conn=conn, noCommit=True, temp=True,
+    upload(tableName,
+           arrays,
+           names,
+           conn=conn,
+           noCommit=True,
+           temp=True,
            analyze=True)
-    res = get(query, conn=conn, preamb=preamb, strLength=strLength,
+    res = get(query,
+              conn=conn,
+              preamb=preamb,
+              strLength=strLength,
               asDict=asDict)
     conn.rollback()
 
